﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Linq;
using System.Text;

namespace PsiKeepers
{
    public enum CharacterType
    {
        Player,
        Enemy,
        EnemyPosessed,
    }

    public class Character
    {
        public bool IsDead { get; set; }

        public PointF Pos { get; set; }

        private float m_angle = 0.0f;
        public float Angle 
        {
            get { return m_angle; }
            set { m_angle = MathUtil.WrapRange(value, 0, MathUtil.PI2);  }
        }

        public PointF Dir
        {
            get
            {
                return PointFHelper.AngleToDir(Angle);
            }
        }

        private CharacterType m_char_type = CharacterType.Enemy;
        public CharacterType Type
        {
            get { return m_char_type; }
            set
            {
                m_char_type = value;
                switch (m_char_type)
                {
                    case CharacterType.Player:
                        m_brush_torso = Brushes.White;
                        m_brush_head = Brushes.Green;
                        m_brush_arms = Brushes.Gray;
                        break;

                    default:
                    case CharacterType.Enemy:
                        m_brush_torso = Brushes.Pink;
                        m_brush_head = Brushes.Red;
                        m_brush_arms = Brushes.DarkOliveGreen;
                        break;

                    case CharacterType.EnemyPosessed:
                        m_brush_torso = Brushes.Pink;
                        m_brush_head = Brushes.Green;
                        m_brush_arms = Brushes.DarkOliveGreen;
                        break;
                }
            }
        }

        private float MoveSpeed
        {
            get
            {
                switch (m_char_type)
                {
                    case CharacterType.Player:
                        return 80.0f;

                    case CharacterType.Enemy:
                        return 40.0f;

                    case CharacterType.EnemyPosessed:
                        return 120.0f;

                    default:
                        throw new Exception();
                }
            }
        }

        private float m_arm_flail_angle = (float)Math.PI / 2.0f;

        private Matrix m_matrix = new Matrix();
        private Brush m_brush_torso;
        private Brush m_brush_head;
        private Brush m_brush_arms;

        public Character(CharacterType type, PointF pos)
        {
            Type = type;
            Pos = pos;
        }

        public void Update(
            float delta_t,
            Controls controls,
            List<Character> chars)
        {
            if (IsDead)
                return;

            // Handle moving
            PointF move_dir = PointF.Empty;
            PointF look_at_dir = PointF.Empty;

            switch (Type)
            {
                case CharacterType.Player:
                    if (controls.plr_move_left)
                        move_dir.X -= 1.0f;
                    if (controls.plr_move_right)
                        move_dir.X += 1.0f;
                    if (controls.plr_move_up)
                        move_dir.Y -= 1.0f;
                    if (controls.plr_move_down)
                        move_dir.Y += 1.0f;
                    break;

                case CharacterType.Enemy:
                    {
                        // Move towards the player
                        Character plr = chars[0];
                        if (plr.Type == CharacterType.Player)
                        {
                            move_dir = plr.Pos.Sub(Pos).Mul(5.0f);
                            look_at_dir = move_dir;
                        }

                        // Repulsion to other characters
                        foreach (Character ch in chars)
                        {
                            if (ch == this)
                                continue;
                            if (ch.Type != CharacterType.Enemy)
                                continue;

                            PointF to_ch = ch.Pos.Sub(Pos);
                            float dist;
                            to_ch = to_ch.UnitOrZeroAndDist(out dist);
                            if (dist < 0.01f)
                                continue;

                            move_dir = move_dir.AddMul(to_ch, -(50000.0f / (dist * dist)));
                        }
                    }
                    break;

                case CharacterType.EnemyPosessed:
                    move_dir = controls.mouse_pos.Sub(Pos);
                    if (move_dir.Abs() < 3.0f)
                        move_dir = PointF.Empty;
                    look_at_dir = move_dir;
                    break;
            }
            Move(delta_t, move_dir, look_at_dir);

            // Avoid other chars
            float avoid_dist = 7.0f * 2.0f;
            float avoid_dist2 = avoid_dist * avoid_dist;
            foreach (Character ch in chars)
            {
                if (ch == this)
                    continue;

                PointF to_ch = ch.Pos.Sub(Pos);
                float dist2 = to_ch.Norm();
                if (dist2 >= avoid_dist2)
                    continue;

                // Move apart
                float change_amt = (avoid_dist2 / (dist2 + avoid_dist2)) - 0.5f;
                to_ch = to_ch.Mul(change_amt);
                Pos = Pos.Sub(to_ch);

                // Explosion?
                if ((Type == CharacterType.Enemy && ch.Type != CharacterType.Enemy) ||
                    (Type != CharacterType.Enemy && ch.Type == CharacterType.Enemy))
                {
                    IsDead = true;
                    ch.IsDead = true;
                }
            }

            // Clamp to screen
            Pos = Pos.ClampToRect(PointF.Empty, Game.ScreenSize);
        }

        public void Draw(Graphics g)
        {
            // Set the transform matrix
            m_matrix.Reset();
            m_matrix.Rotate(MathUtil.Rad2Deg(Angle));
            m_matrix.Translate(Pos.X, Pos.Y, MatrixOrder.Append);
            g.Transform = m_matrix;

            // Draw the character
            float arm_ofs = (float)Math.Cos(m_arm_flail_angle) * 1.5f;
            g.FillRectangle(m_brush_torso, -3, -7, 6, 14);
            g.FillRectangle(m_brush_head, -1, -2, 6, 4);
            g.FillRectangle(m_brush_arms, -4 + arm_ofs, -8, 10, 4);
            g.FillRectangle(m_brush_arms, -4 - arm_ofs, 4, 10, 4);

            //g.DrawLine(new Pen(Brushes.White), 0, 0, 10, 0);

            // Reset the transform matrix
            g.ResetTransform();
        }

        private void Move(float delta_t, PointF dir, PointF look_at_dir)
        {
            // Moving
            if (dir.Norm() > 0.01f)
            {
                float move_speed = delta_t * MoveSpeed;

                dir = dir.Unit();
                Pos = Pos.AddMul(dir, move_speed);

                m_arm_flail_angle += move_speed * 0.2f;

                // Turn towards moving?
                if (look_at_dir.Norm() < 0.01f)
                    look_at_dir = dir;
            }

            // Turning
            if (look_at_dir.Norm() > 0.01f)
            {
                Angle += MathUtil.AngleDiff(Angle, look_at_dir.ToAngle()) * 0.3f;
            }
        }
    }
}
