﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using Microsoft.Win32;

namespace PsiKeepers
{
    public class Game
    {
        public static PointF ScreenSize = PointF.Empty;

        private enum Mode
        {
            Menu,
            Game,
        }

        private class VisualCircle
        {
            public int color;
            public float alpha;
            public PointF pos;
            public float radius;
            public float radius_dest;

            public VisualCircle(PointF _pos, float rad, int r, int g, int b, int a)
            {
                color = (r << 16) | (g << 8) | (b);
                alpha = (float)a / 255.0f;
                pos = _pos;
                radius = 0.0f;
                radius_dest = rad;
            }
        }

        private struct Particle
        {
            public int color;
            public PointF pos;
            public PointF vel;
        }

        private Mode m_mode;

        private Character m_player;
        private List<Character> m_chars = new List<Character>();

        private int m_score;
        private int m_score_best;

        private bool m_psi_active;
        private float m_psi_blink_timer;

        private float m_spawn_timer;
        private float m_spawn_time;

        private float m_psi_gfx_angle;

        private float m_score_alpha;
        private float m_menu_alpha;

        private Background m_background;
        private List<VisualCircle> m_visual_circles = new List<VisualCircle>();
        private List<Particle> m_particles = new List<Particle>();

        private Sounds m_sounds;

        private Random m_rand = new Random();
        private Font m_font_score;
        private Font m_font_best;
        private Font m_font_title;
        private Font m_font_menu;

        private float PsiRadius
        {
            get { return 30.0f; }
        }

        private bool GameActive
        {
            get { return (!m_player.IsDead) && (m_mode == Mode.Game); }
        }

        public Game()
        {
            // Try to read the best score
            try
            {
                RegistryKey key = Registry.CurrentUser.OpenSubKey("Software\\PsiKeepers");
                string score_str = (string)key.GetValue("Hiscore");
                m_score_best = int.Parse(score_str);
            }
            catch
            { // Ignore all loading errors
            }

            // Read the sounds
            m_sounds = new Sounds();

            // Create the fonts
            string font_name = "Comic Sans MS";
            m_font_score = new Font(font_name, 20.0f);
            m_font_best = new Font(font_name, 10.0f);
            m_font_title = new Font(font_name, 50.0f);
            m_font_menu = new Font(font_name, 15.0f);

            // Create the background
            m_background = new Background(ScreenSize);

            // Start the menu
            //StartNew();
            m_player = new Character(CharacterType.Player, PointF.Empty);
            m_player.IsDead = true;

            MoveToMenu();
            m_menu_time_in_menu = 100.0f;
        }

        public void Update(float delta_t, Controls controls)
        {
            // If player is dead, then move to the menu mode
            if (m_player.IsDead && m_mode != Mode.Menu)
                MoveToMenu();

            // Update the menu
            if (m_mode == Mode.Menu)
                UpdateMenu(delta_t, controls);

            // Spawn timer
            if (GameActive)
            {
                m_spawn_timer += delta_t;
                while (m_spawn_timer >= m_spawn_time)
                {
                    m_spawn_timer -= m_spawn_time;
                    m_spawn_time = Math.Max(m_spawn_time * 0.99f, 0.05f);

                    // Pick a new random spawning position
                    PointF pt = PointF.Empty;
                    do
                    {
                        if (m_rand.Next(2) == 0)
                        {
                            pt.X = (float)m_rand.NextDouble() * ScreenSize.X;
                            pt.Y = (float)m_rand.Next(2) * ScreenSize.Y;
                        }
                        else
                        {
                            pt.X = (float)m_rand.Next(2) * ScreenSize.X;
                            pt.Y = (float)m_rand.NextDouble() * ScreenSize.Y;
                        }
                    } while (pt.DistanceTo(m_player.Pos) < 50.0f);

                    // Add the character
                    m_chars.Add(new Character(CharacterType.Enemy, pt));
                }
            }

            // Changing PSI states?
            bool is_psi_active = (GameActive) && controls.mouse_left;
            if (m_psi_active != is_psi_active)
            {
                m_psi_active = is_psi_active;

                if (m_psi_active)
                { // Activate the PSI mode
                    Character nearest = null;
                    float nearest_dist = 0.0f;
                    float psi_rad = PsiRadius;

                    foreach (Character c in m_chars)
                    {
                        if (c.Type != CharacterType.Enemy)
                            continue;

                        float dist = c.Pos.DistanceTo(controls.mouse_pos);
                        if (dist > psi_rad)
                            continue;

                        if (nearest == null ||
                            dist < nearest_dist)
                        {
                            nearest = c;
                            nearest_dist = dist;
                        }
                    }

                    // Found a selected enemy?
                    if (nearest != null)
                    {
                        nearest.Type = CharacterType.EnemyPosessed;

                        m_psi_blink_timer = 0.0f;
                    }
                    else
                    {
                        m_visual_circles.Add(new VisualCircle(controls.mouse_pos, PsiRadius, 255, 100, 100, 220));
                        m_sounds.sound_misclick.Play();
                    }
                }
                else
                { // Deacticate the PSI mode
                    foreach (Character c in m_chars)
                    {
                        if (c.Type == CharacterType.EnemyPosessed)
                            c.Type = CharacterType.Enemy;
                    }
                }
            }

            // Update the PSI blibnking
            m_psi_blink_timer -= delta_t;
            bool do_psi_blink = (m_psi_blink_timer <= 0.0f);
            if (do_psi_blink)
                m_psi_blink_timer = 0.2f;

            // Update the characters
            m_background.BeginUpdate();
            {
                int idx = 0;
                while (idx < m_chars.Count)
                {
                    Character ch = m_chars[idx];

                    // Update the character
                    ch.Update(delta_t, controls, m_chars);

                    // Add to the background
                    m_background.AddChar(ch.Pos);

                    // PSI blink?
                    if (do_psi_blink &&
                        ch.Type == CharacterType.EnemyPosessed)
                    {
                        m_visual_circles.Add(new VisualCircle(ch.Pos, 20.0f, 105, 230, 255, 150));
                        m_sounds.sound_under_control.Play();
                    }

                    // Character died?
                    if (ch.IsDead)
                    {
                        // Give score if a non-player dies
                        if (GameActive && ch.Type != CharacterType.Player)
                        {
                            ++m_score;
                            if (m_score > m_score_best)
                                m_score_best = m_score;
                        }

                        // Add some death particles
                        int pleft = m_rand.Next(50, 80);
                        while (pleft-- > 0)
                        {
                            Particle part = new Particle();
                            int r = m_rand.Next(230, 255);
                            int g = m_rand.Next(180, 205);
                            int b = m_rand.Next(100, 115);
                            int a = m_rand.Next(100, 225);

                            float ang = (float)(m_rand.NextDouble() * 2.0 * Math.PI);
                            PointF dir = PointFHelper.AngleToDir(ang);

                            part.color = (a << 24) | (r << 16) | (g << 8) | (b);
                            part.pos = ch.Pos;
                            part.vel = dir.Mul(MathUtil.Lerp(2.0f, 60.0f, (float)m_rand.NextDouble()));

                            m_particles.Add(part);
                        }

                        // Play a sound
                        switch (m_rand.Next(4))
                        {
                            case 0: m_sounds.sound_expl1.Play(); break;
                            case 1: m_sounds.sound_expl2.Play(); break;
                            case 2: m_sounds.sound_expl3.Play(); break;
                            case 3: m_sounds.sound_expl4.Play(); break;
                        }

                        // Remove the character
                        m_chars.RemoveAt(idx);
                    }
                    else
                        ++idx;
                }
            }
            m_background.EndUpdate();

            // If the player is dead, kill the rest of the chars one by one
            if (!GameActive &&
                m_chars.Count > 0 &&
                m_rand.NextDouble() <= 0.5)
            {
                m_chars[0].IsDead = true;
            }

            // Update the visual circles
            {
                int idx = 0;
                while (idx < m_visual_circles.Count)
                {
                    VisualCircle circle = m_visual_circles[idx];
                    circle.radius = MathUtil.Lerp(circle.radius, circle.radius_dest, 0.3f);
                    if ((circle.radius / circle.radius_dest) > 0.99f)
                        m_visual_circles.RemoveAt(idx);
                    else
                        ++idx;
                }
            }

            // Update the particles
            {
                int idx = 0;
                while (idx < m_particles.Count)
                {
                    Particle part = m_particles[idx];

                    // Update the position
                    part.vel = part.vel.Mul(0.99f);
                    part.pos = part.pos.AddMul(part.vel, delta_t);

                    // Update the color
                    int a = (part.color >> 24) & 0xff;
                    a = (int)((float)a * 0.9f);

                    part.color = (part.color & 0xffffff) | (a << 24);

                    // Either update or remove the particle
                    if (a < 5)
                        m_particles.RemoveAt(idx);
                    else
                    {
                        m_particles[idx] = part;
                        ++idx;
                    }
                }
            }

            // Update the PSI gfx
            m_psi_gfx_angle = MathUtil.WrapRange(m_psi_gfx_angle + delta_t * 200.0f, 0.0f, 360.0f);

            // Update the menu alphas
            if (m_mode == Mode.Game)
            {
                m_score_alpha = MathUtil.Lerp(m_score_alpha, 1.0f, 0.1f);
                m_menu_alpha = MathUtil.Lerp(m_menu_alpha, 0.0f, 0.1f);
            }
            else
            {
                m_score_alpha = MathUtil.Lerp(m_score_alpha, 0.3f, 0.05f);
                m_menu_alpha = MathUtil.Lerp(m_menu_alpha, 1.0f, 0.05f);
            }
        }

        public void Draw(Graphics g, Controls controls)
        {
            // Draw the background
            m_background.Draw(g);

            // Draw the characters
            for (int i = m_chars.Count - 1; i >= 0; --i)
            {
                m_chars[i].Draw(g);
            }

            // Draw the PSI area
            if (!m_player.IsDead &&
                !m_psi_active)
            {
                Pen pen = new Pen(Color.FromArgb(60, 140, 195, 220), 3.0f);

                float rad = PsiRadius;
                RectangleF rect = new RectangleF(controls.mouse_pos.X - rad, controls.mouse_pos.Y - rad, rad * 2.0f, rad * 2.0f);

                for (int i = 0; i < 32; ++i)
                {
                    float ang = m_psi_gfx_angle + (float)i * 3.0f;
                    g.DrawArc(pen, rect, ang, 30.0f);
                    g.DrawArc(pen, rect, ang + 180.0f, 30.0f);
                }
            }

            // Draw the particles
            {
                SolidBrush part_brush = new SolidBrush(Color.White);
                foreach (Particle part in m_particles)
                {
                    part_brush.Color = Color.FromArgb(part.color);
                    g.FillRectangle(part_brush, part.pos.X - 1.0f, part.pos.Y - 1.0f, 3.0f, 3.0f);
                }
            }

            // Draw the visual circles
            foreach (VisualCircle circle in m_visual_circles)
            {
                float r = circle.radius;

                float percent_done = circle.radius / circle.radius_dest;
                float alpha = 1.0f - (float)Math.Pow(percent_done, 3.0);
                int a = (int)(alpha * circle.alpha * 255.0f);

                Pen pen = new Pen(Color.FromArgb(circle.color | (a << 24)), 2.0f);
                g.DrawEllipse(pen, circle.pos.X - r, circle.pos.Y - r, r * 2.0f, r * 2.0f);
            }

            // Draw the score
            {
                Brush score_brush = new SolidBrush(Color.FromArgb((int)(m_score_alpha * 255.0f), 90, 255, 170));

                g.DrawString("Score: " + m_score, m_font_score, score_brush, 10, 10);
                g.DrawString("Best: " + m_score_best, m_font_best, score_brush, 15, 10 + m_font_score.Height);
            }

            // Draw the menu
            if (m_menu_alpha > 0.02f)
            {
                Brush title_brush = new SolidBrush(Color.FromArgb((int)(m_menu_alpha * 255.0f), 90, 255, 70));
                Brush menu_brush = new SolidBrush(Color.FromArgb((int)(m_menu_alpha * 255.0f), 140, 225, 100));

                WriteCentered(g, "Psi Keepers", title_brush, 20, m_font_title);

                int y = 160;
                int h = m_font_menu.Height;
                WriteCentered(g, "Use WASD or arrow keys to move.", menu_brush, y, m_font_menu); y += h;
                WriteCentered(g, "Avoid the suicide bombers and try not to die.", menu_brush, y, m_font_menu); y += h;
                y += h / 2;
                WriteCentered(g, "Holding the left mouse button down allows you to", menu_brush, y, m_font_menu); y += h;
                WriteCentered(g, "control the suicide bomber closest to the PSI circle.", menu_brush, y, m_font_menu); y += h;
                y += h / 2;
                WriteCentered(g, "Bombers under control follow the mouse pointer;", menu_brush, y, m_font_menu); y += h;
                WriteCentered(g, "run the bombers into each other.", menu_brush, y, m_font_menu); y += h;

                y = (int)ScreenSize.Y - h * 2;
                WriteCentered(g, "Press left mouse button to start", menu_brush, y, m_font_menu); y += h;
            }
        }

        private void WriteCentered(Graphics g, string txt, Brush brush, float y, Font font)
        {
            SizeF txt_size = g.MeasureString(txt, font);
            g.DrawString(txt, font, brush, (ScreenSize.X - txt_size.Width) * 0.5f, y);
        }

        private void StartNew()
        {
            m_mode = Mode.Game;

            m_chars.Clear();

            m_player = new Character(CharacterType.Player, ScreenSize.Mul(0.5f));
            m_chars.Add(m_player);

            m_score = 0;

            m_psi_active = false;
            m_psi_blink_timer = 0.0f;
            m_spawn_time = 0.8f;
        }

        /////////////////////////////////////////////

        private bool m_menu_left_down;
        private float m_menu_time_in_menu;

        private void MoveToMenu()
        {
            m_mode = Mode.Menu;
            m_menu_left_down = false;
            m_menu_time_in_menu = 0.0f;

            // Try to save the best score
            try
            {
                RegistryKey key = Registry.CurrentUser.CreateSubKey("Software\\PsiKeepers");
                key.SetValue("Hiscore", m_score_best.ToString());
                key.Flush();
                key.Close();
            }
            catch
            { // Ignore all saving errors
            }
        }

        private void UpdateMenu(float delta_t, Controls controls)
        {
            m_menu_time_in_menu += delta_t;

            // Track starting the menu
            if (controls.mouse_left == false &&
                m_menu_left_down &&
                m_menu_time_in_menu > 1.0f)
            {
                m_sounds.sound_click.Play();
                StartNew();
                return;
            }
            m_menu_left_down = controls.mouse_left;
        }
    }
}
